/*
xpopclk Version 0.7.0 - World Population Clock
for X Windows.
Copyright (C) 2001,2002  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* to define the escape key */
#define XK_MISCELLANY 1

#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <sys/times.h>
#include <X11/Xlib.h>
#include <X11/keysymdef.h>
#include <assert.h>
#include <unistd.h>

#define FONT "linux8x16"
#define NIL (0)
#define SCRW 240
#define SCRH 120

unsigned int dayticks;
double totpop;
double dbltkdlta;
time_t origclk;

int whiteColor,blackColor;
int runflg;
Display *dpy;
Window w;
GC gc;
Font fontid;

void putstx(pgm)
char *pgm;
   {
   fprintf(stderr,"Usage: %s <annual_change>\n",
      pgm);
   fprintf(stderr,"Where annual_change is the "
      "annual increase or decrease\n");
   fprintf(stderr,"from -999999999999999 to "
      "999999999999999\n");
   exit(1);
   } /* putstx */

void slp()
   {
   int rslt;
   struct timespec request;
   struct timespec remain;
   request.tv_sec  = (long int) 0;
   request.tv_nsec = (long int) 1;
   rslt = nanosleep(&request,&remain);
   if (rslt < 0)
      {
      fprintf(stderr,"Nanosleep error %d\n",
         errno);
      perror("Nanosleep error");
      exit(1);
      } /* nanosleep error */
   } /* slp using nanosleep */

/* X Windows code is based on: */
/* http://tronche.lri.fr:8000/gui/x/xlib-tutorial/2nd-program-anatomy.html */

void getkey()
   {
   int i,rslt,msk;
   int symbol;
   int XCheckMaskEvent();
   XEvent e;
   XKeyEvent *k;

   msk = KeyPressMask|ExposureMask;

   XSelectInput(dpy, w, msk);

   while (XCheckMaskEvent(dpy, msk, &e))
      {
      if (e.type == KeyPress)
         {
         k = (XKeyEvent *) &e;
         symbol = XLookupKeysym(k,0);
         if (symbol == XK_Escape)
            {
            runflg = 0;
            } /* if quit */
         } /* if keypress event */
      else if (e.type == Expose)
         {
         XDrawImageString(dpy,w,gc,30,40,
            "World Population Change", 23);
         } /* if expose event */
      } /* if event received */
   } /* getkey */

void initx()
   {
   int i,rslt;
   char title[64];

   dpy = XOpenDisplay(NIL);

   if (dpy == NULL)
      {
      fprintf(stderr,"X Windows failure\n");
      exit(1);
      } /* if X Windows is not active */

   assert(dpy);

   whiteColor = WhitePixel(dpy, DefaultScreen(dpy));
   blackColor = BlackPixel(dpy, DefaultScreen(dpy));

   w = XCreateSimpleWindow(dpy,
      DefaultRootWindow(dpy),
      0, 0, 
      SCRW, SCRH,
      0, whiteColor,
      whiteColor);

   XSelectInput(dpy, w, StructureNotifyMask);

   XMapWindow(dpy, w);

   gc = XCreateGC(dpy, w, 0, NIL);

   fontid = (Font) XLoadFont(dpy,FONT);

   XSetFont(dpy,gc,fontid);

   XSetForeground(dpy, gc, blackColor);

   XSetWindowBorderWidth(dpy, w, 40);

   sprintf(title,"xpopclk");
   XStoreName(dpy,w,title);

   XMoveWindow(dpy,w,0,0);

   while(1)
      {
      XEvent e;
      XNextEvent(dpy, &e);
      if (e.type == MapNotify) break;
      } /* wait for window initialization */
   } /* initx */

int gettps()
   {
   unsigned int diff;
   time_t frstsec,currsec;
   clock_t prev,clk;
   struct tms t;
   /* get first date & time in seconds */
   time(&frstsec);
   currsec = frstsec;
   while (currsec == frstsec)
      {
      /* get first date & time in seconds */
      time(&currsec);
      slp();
      } /* loop until second starts */
   /* get first clock ticks since boot */
   prev = times(&t);
   frstsec = currsec;
   while (currsec == frstsec)
      {
      /* get current date & time in seconds */
      time(&currsec);
      slp();
      } /* loop for 1 second */
   /* get clock ticks since boot */
   clk = times(&t);
   if (clk < prev)
      {
      fprintf(stderr,"Negative ticks, rerun job\n");
      exit(0);
      }
   diff = clk - prev;
   printf("Ticks per second %d\n", diff);
   return(diff);
   } /* gettps */

void count()
   {
   unsigned int diff;
   double tgtpop,currpop;
   time_t clk;
   struct tms t;
   char str[64];
   currpop = totpop;
   if (dbltkdlta < 0.0)
      tgtpop = totpop - 1.0;
   else
      tgtpop = totpop + 1.0;
   while (runflg)
      {
      if (dbltkdlta < 0.0
         && currpop < tgtpop) break;
      else if (dbltkdlta >= 0.0
         && currpop > tgtpop) break;
      clk = times(&t);
      if (clk < origclk)
         diff = clk + dayticks - origclk;
      else diff = clk - origclk;
      currpop = dbltkdlta * diff;
      if ((dbltkdlta < 0.0
         && currpop > tgtpop)
         || (dbltkdlta >= 0.0
         && currpop < tgtpop))
         {
         getkey();
         } /* if sleep every second */
      if (runflg) slp();
      } /* wait for birth */
   totpop = currpop;
   sprintf(str,"%10.0f", totpop);
   XDrawImageString(dpy,w,gc,50,70,str,strlen(str));
   } /* count */

void runclk()
   {
   while (runflg)
      {
      count();
      getkey();
      if (runflg) slp();
      } /* display loop */
   } /* runclk */

int main(argc,argv)
int argc;
char **argv;
   {
   int tps;
   double dblanndlta,dblspy,sign;
   double dbltps,dbltpy;
   double ten,unit;
   char *p;
   struct tms t;
   if (argc != 2) putstx(*argv);
   p = *(argv+1);
   if (*p == '-')
      {
      sign = -1.0;
      p++;
      }
   else sign = 1.0;
   dblanndlta = 0.0;
   ten = 10.0;
   while (*p != '\0')
      {
      if (*p < '0' || *p > '9')
         putstx(*argv);
      unit = (double) *p - '0';
      dblanndlta = (dblanndlta * ten) + unit;
      p++;
      }
   if (dblanndlta > 999999999999999)
      putstx(*argv);
   /* Initialize X Windows */
   initx();
   /* calculate constants */
   /* annual change in population */
   dblanndlta *= sign;
   /* seconds per year */
   dblspy = 3600.0 * 24.0 * 365.25;
   /* ticks per second */
   tps = gettps();
   if (tps < 1 || tps > 999999)
      {
      fprintf(stderr,"Invalid tps %d\n", tps);
      exit(1);
      } /* if tps error */
   dbltps = (double) tps;
   /* ticks per year */
   dbltpy = dblspy * dbltps;
   /* population change per tick */
   dbltkdlta = dblanndlta / dbltpy;
   /* ticks per day */
   dayticks = 3600 * 24 * tps;
   XDrawImageString(dpy,w,gc,30,40,
      "World Population Change", 23);
   /* start counting now */
   origclk = times(&t);
   totpop = 0.0;
   runflg = 1;
   runclk();
   return(0);
   } /* main */
